"""Pure-Python reimplementation of the playout code.

This is for timing comparison.

"""

import time

import pyego
from pyego import BLACK, WHITE, Pass, Any

class Playout(object):

    def __init__(self, board, first_player):
        self.board = board
        self.act_player = first_player
        self.last_v = {BLACK : Any, WHITE : Any}
        self.move_no = 0

    def get_last_play(self, player):
        return self.last_v[player]

    def play_one(self):
        board = self.board
        act_player = self.act_player
        start = pyego.playout_rand_int(board.empty_v_cnt)
        for rng in ((start, board.empty_v_cnt),
                    (0, start)):
            for i in xrange(*rng):
                v = board.get_empty_vertex(i)
                if (not (board.is_eyelike(act_player, v)) and
                    board.play_internal(act_player, v) < pyego.PLAY_SS_SUICIDE):
                    self.last_v[act_player] = v
                    return
        self.last_v[act_player] = Pass

    def run(self):
        while True:
            self.play_one()
            self.act_player = pyego.other_player(self.act_player)
            self.move_no += 1
            if self.last_v[BLACK] == self.last_v[WHITE] == Pass:
                return pyego.PLAYOUT_OK
            if abs(self.board.approx_score()) > pyego.MERCY_THRESHOLD:
                return pyego.PLAYOUT_MERCY
            if self.move_no >= pyego.MAX_PLAYOUT_LENGTH:
                return pyego.PLAYOUT_TOO_LONG

    def run_fast(self):
        board = self.board
        act_player = self.act_player
        MERCY_THRESHOLD = pyego.MERCY_THRESHOLD
        MAX_PLAYOUT_LENGTH = pyego.MAX_PLAYOUT_LENGTH
        PLAY_SS_SUICIDE = pyego.PLAY_SS_SUICIDE
        get_empty_vertex = board.get_empty_vertex
        is_eyelike = board.is_eyelike
        play_internal = board.play_internal
        approx_score = board.approx_score
        playout_rand_int = pyego.playout_rand_int
        other_player = pyego.other_player
        abs_ = abs
        last_v = [Any, Any]
        move_no = 0
        any_pass = 0
        while 1:
            end = board.empty_v_cnt
            i = start = playout_rand_int(end)
            while 1:
                v = get_empty_vertex(i)
                if (not (is_eyelike(act_player, v)) and
                    play_internal(act_player, v) < PLAY_SS_SUICIDE):
                    last_v[act_player] = v
                    break
                i = i + 1
                if i == end:
                    i = 0
                if i == start:
                    any_pass = 1
                    last_v[act_player] = Pass
                    break
            if any_pass and last_v[BLACK] == last_v[WHITE] == Pass:
                return pyego.PLAYOUT_OK
            if abs_(approx_score()) > MERCY_THRESHOLD:
                return pyego.PLAYOUT_MERCY
            move_no = move_no + 1
            if move_no >= MAX_PLAYOUT_LENGTH:
                return pyego.PLAYOUT_TOO_LONG
            act_player = other_player(act_player)


def run_playout_benchmark(start_board, playout_count, first_player,
                          use_run_fast=False, use_native_playout=False):
    wins = {BLACK : 0, WHITE : 0}

    if use_native_playout:
        Playout_cls = pyego.Playout
        runner = Playout_cls.run
        assert not use_run_fast
    else:
        Playout_cls = Playout
        if use_run_fast:
            runner = Playout_cls.run_fast
        else:
            runner = Playout_cls.run

    seconds_begin = time.clock()

    for i in xrange(playout_count):
        working_board = start_board.copy()
        playout = Playout_cls(working_board, first_player)
        status = runner(playout)
        if status == pyego.PLAYOUT_OK:
            wins[working_board.winner()] += 1
        elif status == pyego.PLAYOUT_MERCY:
            wins[working_board.approx_winner()] += 1

    seconds_end = time.clock()
    seconds_total = seconds_end - seconds_begin;
    p_black_win = float(wins[BLACK]) / float(wins[BLACK] + wins[WHITE])

    return ("Initial board:\n" +
            "komi %s\n%s" % (start_board.get_komi(), start_board.to_string()) +
            "Performance: \n" +
            "  %s playouts\n" % playout_count +
            "  %f seconds\n" % seconds_total +
            "  %f kpps\n" % (playout_count / seconds_total / 1000.0) +
            "Black wins = %s\n" % wins[BLACK] +
            "White wins = %s\n" % wins[WHITE] +
            "P(black win) = %f\n" % p_black_win)

def main():
    import sys
    if '--psyco' in sys.argv:
        import psyco
        psyco.full()
    use_run_fast = ('--fast' in sys.argv)
    use_native_playout = ('--native' in sys.argv)
    b = pyego.Board()
    print run_playout_benchmark(b, 6000, BLACK,
                                use_run_fast, use_native_playout)

if __name__ == "__main__":
    main()
