"""Simple tests for pyego.

Run this with py.test (see http://codespeak.net/).

"""

RUN_SLOW_TESTS = False

from py.test import raises

import pyego
from pyego import Vertex, BLACK, WHITE, EMPTY

def test_vertex():
    v = Vertex(2, 3)
    assert v.row == 2
    assert v.col == 3
    assert v.coords == (2, 3)
    raises(ValueError, Vertex, 3, 11)
    raises(ValueError, Vertex, -1, 5)
    raises(ValueError, Vertex, 5, -1)
    raises(TypeError, Vertex)
    v1 = Vertex(4, 4)
    v2 = Vertex(4, 4)
    v3 = Vertex(0, 0)
    assert v1 == v2
    assert not (v1 != v2)
    assert v1 != v3
    assert not (v1 == v3)
    raises(TypeError, "v1 <= v2")
    d = {}
    d[v1] = "a"
    d[v2] = "b"
    d[v3] = "c"
    assert len(d) == 2
    assert d[v1] == "b"

def test_pseudovertices():
    assert pyego.Any != pyego.Pass
    assert pyego.Any != pyego.Resign
    assert pyego.Pass != pyego.Resign
    v = Vertex(0, 0)
    assert pyego.Any != v
    assert pyego.Pass != v
    assert pyego.Resign != v
    assert not (pyego.Any == v)
    assert not (pyego.Pass == v)
    assert not (pyego.Resign == v)
    assert v != pyego.Any
    assert v != pyego.Pass
    assert v != pyego.Resign
    assert not (v == pyego.Any)
    assert not (v == pyego.Pass)
    assert not (v == pyego.Resign)
    assert str(pyego.Any) == "<Pseudovertex any>"
    assert str(pyego.Pass) == "<Pseudovertex pass>"
    assert str(pyego.Resign) == "<Pseudovertex resign>"

def test_vertex_name():
    assert Vertex(0, 0).name == "A9"
    assert Vertex(0, 8).name == "J9"
    assert Vertex(5, 5).name == "F4"
    v = Vertex(0, 0)
    assert str(v) == repr(v) == "<Vertex A9>"

def test_vertex_from_name():
    assert pyego.vertex_from_name("A9").coords == (0, 0)
    assert pyego.vertex_from_name("J9").coords == (0, 8)
    assert pyego.vertex_from_name("F4").coords == (5, 5)
    assert pyego.vertex_from_name("a1").coords == (8, 0)
    assert pyego.vertex_from_name(" A1").coords == (8, 0)
    assert pyego.vertex_from_name("A1 ").coords == (8, 0)
    assert pyego.vertex_from_name("\tA1\t").coords == (8, 0)
    assert pyego.vertex_from_name("\nA1\n").coords == (8, 0)
    assert pyego.vertex_from_name("\fA1\f").coords == (8, 0)
    raises(ValueError, pyego.vertex_from_name, "")
    raises(ValueError, pyego.vertex_from_name, "I5")
    raises(ValueError, pyego.vertex_from_name, "A10")
    raises(ValueError, pyego.vertex_from_name, "resign")
    raises(ValueError, pyego.vertex_from_name, "pass")

def test_board():
    b = pyego.Board()

    raises(ValueError, b.play, 2, Vertex(5, 5))
    raises(TypeError, b.play, BLACK, None)
    assert b.empty_v_cnt == 81

    b.play(WHITE, Vertex(0, 0))
    b.play(BLACK, Vertex(0, 1))
    b.play(WHITE, Vertex(1, 1))
    b.play(WHITE, Vertex(2, 0))
    assert b.empty_v_cnt == 77

    # Point occupied
    raises(pyego.IllegalMove, b.play, BLACK, Vertex(0, 0))

    assert b.to_string() == """\
   A B C D E F G H J
 9 O # . . . . . . . 9
 8 . O . . . . . . . 8
 7 O . . . . . . . . 7
 6 . . . . . . . . . 6
 5 . . . . . . . . . 5
 4 . . . . . . . . . 4
 3 . . . . . . . . . 3
 2 . . . . . . . . . 2
 1 . . . . . . . . . 1
   A B C D E F G H J
"""

    b.play(BLACK, Vertex(1, 0))
    # Forbidden by simple ko
    raises(pyego.IllegalMove, b.play, WHITE, Vertex(0, 0))

    assert b.to_string() == """\
   A B C D E F G H J
 9 . # . . . . . . . 9
 8 # O . . . . . . . 8
 7 O . . . . . . . . 7
 6 . . . . . . . . . 6
 5 . . . . . . . . . 5
 4 . . . . . . . . . 4
 3 . . . . . . . . . 3
 2 . . . . . . . . . 2
 1 . . . . . . . . . 1
   A B C D E F G H J
"""

    assert b.point(Vertex(0, 0)) == EMPTY
    assert b.point(Vertex(0, 1)) == BLACK
    assert b.point(Vertex(1, 1)) == WHITE

    b.set_komi(0.5)
    assert b.score() == 1
    assert b.approx_score() == 0
    assert b.winner() == BLACK
    assert b.approx_winner() == WHITE


POSITION = """\
9
. . . . . . . . .
. . . . . . # O .
. . # . # # O . O
. # . # . . O O .
. # # O O O O . .
. # O O . . . . .
# O O . . # . O .
. # O . # O # # .
. . # O . O . # .
"""

BAD_POSITION = """\
9
. . . . . . . . .
. . . . . . # O .
. . X . # # O . O
. # . # . . O O .
. # # O O O O . .
. # O O . . . . .
# O O . . # . O .
. # O . # O # # .
. . # O . O . # .
"""

VERY_BAD_POSITION = """\
9
. . . . . . . . .
. . . . . . # O .
. . # . # # O # O
. # . # . . O O .
. # # O O O O . .
. # O O . . . . .
# O O . . # . O .
. # O . # O # # .
. . # O . O . # .
"""

RESULT = """\
   A B C D E F G H J
 9 . . . . . . . . . 9
 8 . . . . . . # O . 8
 7 . . # . # # O . O 7
 6 . # . # . . O O . 6
 5 . # # O O O O . . 5
 4 . # O O . . . . . 4
 3 # O O . . # . O . 3
 2 . # O . # O # # . 2
 1 . . # O . O . # . 1
   A B C D E F G H J
"""

CLEAR = """\
   A B C D E F G H J
 9 . . . . . . . . . 9
 8 . . . . . . . . . 8
 7 . . . . . . . . . 7
 6 . . . . . . . . . 6
 5 . . . . . . . . . 5
 4 . . . . . . . . . 4
 3 . . . . . . . . . 3
 2 . . . . . . . . . 2
 1 . . . . . . . . . 1
   A B C D E F G H J
"""

def test_load():
    b = pyego.Board()
    b.load(POSITION)
    assert b.to_string() == RESULT
    raises(ValueError, b.load, BAD_POSITION)
    assert b.to_string() == CLEAR
    # This will exit the process.
    #b.load(VERY_BAD_POSITION)


EYES_POSITION = """\
9
. # # # O O O . O
# . # . # O . O O
. # # # # O O O #
# . # . # O . O #
. # # # # # O . #
# O O O # O . O #
# # O . O . O . #
. # # O . O O O #
# # O . O . O # .
"""

EYES_VALUES = """\
b # # # O O O w O
# b # b # O w O O
b # # # # O O O #
# b # b # O w O #
. # # # # # O . #
# O O O # O w O #
# # O . O w O . #
b # # O w O O O #
# # O . O w O # .
"""

def test_is_eyelike():
    b = pyego.Board()
    raises(TypeError, b.is_eyelike, BLACK, None)
    b.load(EYES_POSITION)
    for row, s in enumerate(EYES_VALUES.split("\n")):
        for col, c in enumerate(s[::2]):
            if c in ('.bw'):
                assert b.is_eyelike(BLACK, Vertex(row, col)) == (c == 'b')
                assert b.is_eyelike(WHITE, Vertex(row, col)) == (c == 'w')


def test_load_board():
    b1 = pyego.Board()
    b1.load(EYES_POSITION)
    b2 = pyego.Board()
    b2.load_board(b1)
    s = b2.to_string()
    assert s == b1.to_string()
    b1.clear()
    assert b1.to_string() == CLEAR
    assert b2.to_string() == s
    raises(TypeError, b2.load_board, None)


def test_copy():
    b1 = pyego.Board()
    b1.load(EYES_POSITION)
    b2 = b1.copy()
    s = b2.to_string()
    assert s == b1.to_string()
    b1.clear()
    assert b1.to_string() == CLEAR
    assert b2.to_string() == s

def test_stack_board():
    sb = pyego.StackBoard()
    raises(IndexError, sb.try_undo)

    raises(TypeError, sb.try_play, WHITE, None)
    raises(TypeError, sb.try_play, WHITE, "A4")
    raises(TypeError, sb.try_play, WHITE, pyego.Any)
    assert sb.is_hash_repeated() is False
    sb.try_play(WHITE, Vertex(1, 2))
    assert sb.is_hash_repeated() is False
    sb.try_play(WHITE, pyego.Pass)
    assert sb.is_hash_repeated() is True
    sb.try_play(WHITE, pyego.Resign)

    assert sb.is_legal(BLACK, Vertex(1, 3))
    assert not sb.is_legal(BLACK, Vertex(1, 2))
    assert sb.act_board().get_komi() == 7.5
    sb.set_komi(0.5)
    assert sb.act_board().get_komi() == 0.5

    assert sb.act_board().point(Vertex(1, 3)) == EMPTY
    assert sb.act_board().point(Vertex(1, 2)) == WHITE
    sb.try_undo()
    assert sb.act_board().point(Vertex(1, 2)) == WHITE
    sb.try_undo()
    assert sb.act_board().point(Vertex(1, 2)) == WHITE
    sb.try_undo()
    assert sb.act_board().point(Vertex(1, 2)) == EMPTY
    raises(IndexError, sb.try_undo)
    assert sb.is_hash_repeated() is False
    assert sb.act_board().get_komi() == 0.5

def test_stack_board_load():
    sb = pyego.StackBoard()
    sb.try_play(WHITE, Vertex(1, 2))
    sb.try_play(BLACK, Vertex(2, 2))
    raises(ValueError, sb.load, BAD_POSITION)
    assert sb.act_board().point(Vertex(2, 2)) == BLACK
    sb.try_undo()
    assert sb.act_board().point(Vertex(2, 2)) == EMPTY
    assert sb.act_board().point(Vertex(1, 2)) == WHITE
    sb.load(POSITION)
    assert sb.act_board().to_string() == RESULT
    raises(IndexError, sb.try_undo)

PLAYOUT_POSITION = """\
9
. # # # O O O . O
# . # . # O . O O
. # # # # O O O #
# . # . # O . O #
. # # # # # O . #
# O O O # O . O #
# # O . O . O . #
. # # O . O O O #
# # O . O . O # .
"""

PLAYOUT_TERMINAL = """\
   A B C D E F G H J
 9 . # # # O O O . O 9
 8 # . # . # O . O O 8
 7 . # # # # O O O O 7
 6 # . # . # O . O O 6
 5 # # # # # # O O . 5
 4 # O O O # O . O O 4
 3 # # O O O . O O . 3
 2 . # # O . O O O O 2
 1 # # O O O . O O O 1
   A B C D E F G H J
"""

# These tests rely on the particular order of moves chosen by the playout code.
# They will probably break if there are any changes to this code.

def test_random():
    pyego.set_playout_seed(123)
    assert pyego.playout_rand_int(100) == 54
    assert pyego.playout_rand_int(100) == 31

def test_playout():
    pyego.set_playout_seed(123)
    b = pyego.Board()
    b.load(PLAYOUT_POSITION)
    playout = pyego.Playout(b, BLACK)
    assert playout.get_last_play(BLACK) is pyego.Any
    assert playout.get_last_play(WHITE) is pyego.Any
    assert playout.board is b
    passed = False
    moves = 0
    while True:
        playout.play_one()
        move = playout.get_last_play(playout.act_player)
        if move == pyego.Pass:
            if passed:
                break
            passed = True
        else:
            passed = False
        moves += 1
        assert moves < 100
        playout.act_player = pyego.other_player(playout.act_player)
    assert moves == 21
    assert b.to_string() == PLAYOUT_TERMINAL

def test_run():
    pyego.set_playout_seed(123)
    b = pyego.Board()
    b.load(PLAYOUT_POSITION)
    playout = pyego.Playout(b, BLACK)
    assert playout.run() == pyego.PLAYOUT_OK
    playout.play_one()
    assert playout.get_last_play(playout.act_player) == pyego.Pass
    assert b.to_string() == PLAYOUT_TERMINAL

def test_stack_board_act_board():
    sb = pyego.StackBoard()
    b = sb.act_board()
    b.play(WHITE, Vertex(1, 1))
    b2 = sb.act_board()
    assert b.to_string() == b2.to_string()
    s = b.to_string()

    # Try to check that the underlying board is being kept around
    # (it's a bit random whether this fails if it isn't).
    sb = pyego.StackBoard()
    b3 = sb.act_board()
    assert b.to_string() == b2.to_string() == s

def test_uct():
    if RUN_SLOW_TESTS:
        pyego.set_playout_seed(123)
        sb = pyego.StackBoard()
        uct = pyego.Uct(sb)
        del sb
        assert uct.genmove(BLACK) == pyego.vertex_from_name("E3")
