"""Python wrapper for Lukasz Lew's Library of Effective GO routines (libego).

"""

cdef extern from *:
    cdef enum:
        Py_EQ
        Py_NE

cdef extern from "libego_wrapper.cpp":

    ctypedef int bool_ "bool"
    ctypedef struct string: # C++ std::string
        char * (*c_str) ()
    string string_from_cstr "string" (char *c)
    ctypedef struct ostringstream:
        string (*str) ()
    ctypedef int uint64

    cdef enum:
        board_size
        board_area
        max_playout_length
        mercy_threshold

    ctypedef int coord_t
    ctypedef unsigned int uint
    ctypedef unsigned long long uint64

    cdef struct player_t:
        int (*get_idx) ()
        bool_ (*in_range) ()
        player_t (*other) ()
    player_t player_from_int "player_t" (uint idx)

    cdef struct vertex_t:
        int (*row) ()
        int (*col) ()
        int (*get_idx) ()
        bool_ (*is_on_board) ()
        bool_ (*eq "operator==") (vertex_t v)
        string (*to_string) ()
    vertex_t vertex_from_coords "vertex_t" (coord_t r, coord_t c)

    cdef enum color_t "color::t":
        black_color "color::black"
        white_color "color::white"
        empty_color "color::empty"

    cdef struct hash_t:
        uint64 hash
    vertex_t vertex_from_coords "vertex_t" (coord_t r, coord_t c)

    cdef enum play_ret_t:
        play_ok
        play_suicide
        play_ss_suicide
        play_ko

    ctypedef struct vertex_to_color_map_t:
        color_t (*get "operator[]") (vertex_t v)

    cdef enum play_ret_t:
        pass

    ctypedef struct board_t:
        vertex_to_color_map_t color_at
        hash_t hash
        vertex_t empty_v[board_area]
        int empty_v_cnt
        void (*clear) ()
        void (*set_komi) (float fkomi)
        float (*get_komi) ()
        play_ret_t (*play_no_pass) (player_t player, vertex_t v)
        bool_ (*is_eyelike) (player_t player, vertex_t v)
        int (*approx_score) ()
        player_t (*approx_winner) ()
        int (*score) ()
        player_t (*winner) ()
        string (*to_string) ()
        string (*to_string_with_mark "to_string") (vertex_t v)
        void load (board_t *save_board)
    board_t *new_board "new board_t" ()
    void del_board "delete" (board_t *bd)

    ctypedef struct player_to_vertex_map_t:
        vertex_t (*get "operator[]") (player_t pl)

    cdef enum playout_status:
        playout_ok
        playout_mercy
        playout_too_long

    ctypedef struct simple_playout_t:
        player_t act_player
        player_to_vertex_map_t last_v
        void (*play_one) ()
        playout_status (*run) ()
    simple_playout_t *new_simple_playout "new simple_playout_t" (
        board_t *bd, player_t pl)
    void del_simple_playout "delete" (simple_playout_t *plt)

    void _run_simple_playout_benchmark "simple_playout_benchmark::run" (
        board_t *start_board, int playout_cnt, player_t first_player,
        ostringstream out)

    ctypedef struct stack_board_t:
        board_t *(*act_board) ()
        bool_ (*try_play) (player_t player, vertex_t v)
        bool_ (*try_undo) ()
        bool_ (*is_legal) (player_t pl, vertex_t v)
        bool_ (*is_hash_repeated) ()
        void (*set_komi) (float komi)
    stack_board_t *new_stack_board "new stack_board_t" ()
    void del_stack_board "delete" (stack_board_t *bd)

    ctypedef struct uct_t:
        vertex_t (*genmove) (player_t pl)
    uct_t *new_uct "new uct_t" (stack_board_t sb)
    void del_uct "delete" (uct_t *bd)

    ctypedef struct gtp_engine_t:
        pass

    ctypedef struct gtp_static_commands_t:
        void (*add) (string command, string response)
    gtp_static_commands_t *new_gtp_static_commands \
                          "new gtp_static_commands_t" ()
    void del_gtp_static_commands "delete" (gtp_static_commands_t *bd)

    ctypedef struct gtp_board_t:
        pass
    gtp_board_t *new_gtp_board "new gtp_board_t" (stack_board_t sb)
    void del_gtp_board "delete" (gtp_board_t *bd)

    ctypedef struct gtp_genmove_t:
        pass
    gtp_genmove_t *new_gtp_genmove \
                  "new gtp_genmove_t" (stack_board_t st, uct_t uct)
    void del_gtp_genmove "delete" (gtp_genmove_t *bd)

    ctypedef struct gtp_t:
        void (*run_loop) ()
        void (*register_engine) (gtp_engine_t gtp_engine)
    gtp_t *new_gtp "new gtp_t" ()
    void del_gtp "delete" (gtp_t *bd)


cdef extern from "libego_helpers.cpp":

    vertex_t get_vertex_pass()
    vertex_t get_vertex_any()
    vertex_t get_vertex_resign()
    vertex_t vertex_from_alnum(string s)
    int load_board_from_string (board_t b, string s)
    int load_stack_board_from_string (stack_board_t sb, string s)
    void set_playout_seed_impl(unsigned long seed)
    inline unsigned long playout_rand_int_impl(uint n)


cdef extern from "gtp_py.cpp":

    ctypedef struct gtp_py_t:
        pass
    gtp_py_t *new_gtp_py "new gtp_py_t" (pyengine, e1, e2)
    void del_gtp_py "delete" (gtp_py_t *bd)


## Public API

# Black and white have the same numeric values when representing players and
# when representing board points; taking them from the board points.
BLACK = black_color
WHITE = white_color
EMPTY = empty_color

BOARD_SIZE = board_size
BOARD_AREA = board_area
MAX_PLAYOUT_LENGTH = max_playout_length
MERCY_THRESHOLD = mercy_threshold


def color_name(color):
    """Return a string naming the specified colour."""
    return {BLACK : 'black',
            WHITE : 'white',
            EMPTY : 'empty'}[color]

def other_player(int player):
    """Return the value representing the other player."""
    cdef player_t pl = player_from_int(player)
    if not pl.in_range():
        raise ValueError("invalid player")
    return pl.other().get_idx()


cdef vertex_t vertex_pass = get_vertex_pass()
cdef vertex_t vertex_any = get_vertex_any()
cdef vertex_t vertex_resign = get_vertex_resign()


class Pseudovertex(object):
    """Constant values sometimes used in place of vertices."""
    # The only reason to have a class for these is to get a readable repr().
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "<Pseudovertex %s>" % self.name

Pass = Pseudovertex("pass")
Any = Pseudovertex("any")
Resign = Pseudovertex("resign")


cdef class Vertex:
    """A point on the Go board.

    Instantiate with coordinates (row, col), using integers in the range
    0 .. board_size-1. (0, 0) is the top-left of the board (A9 on a 9x9 board).

    Vertex objects immutable (like int or str objects).

    Note that Pass, Any, and Resign are not Vertex objects.

    Read-only attributes:
      row    -- the row coordinate
      col    -- the col coordinate
      coords -- (row, col)
      name   -- The 'A1'-style notation for the point

    """
    # We ensure that this always satisfies is_on_board()
    cdef vertex_t _v

    def __init__(self, int row, int col):
        if not ((0 <= row < board_size) and (0 <= col < board_size)):
            raise ValueError("coordinates out of range")
        self._v = vertex_from_coords(row, col)

    def __richcmp__(Vertex self, object other, int opid):
        if not isinstance(other, Vertex):
            if opid == Py_EQ:
                return False
            elif opid == Py_NE:
                return True
            else:
                raise TypeError
        cdef Vertex oth = other
        if opid == Py_EQ:
            return self._v.get_idx() == oth._v.get_idx()
        elif opid == Py_NE:
            return self._v.get_idx() != oth._v.get_idx()
        else:
            # Or just order them?
            raise TypeError("vertices are not ordered")

    def __hash__(Vertex self):
        # Can't be negative, therefore can't be -1
        return self._v.get_idx()

    property row:
        def __get__(self):
            return self._v.row()

    property col:
        def __get__(self):
            return self._v.col()

    property coords:
        def __get__(self):
            return (self._v.row(), self._v.col())

    property name:
        def __get__(self):
            return self._v.to_string().c_str()

    def __repr__(self):
        return "<Vertex %s>" % self.name

cdef Vertex _vertex_from_vertex_t(vertex_t v):
    cdef Vertex result = Vertex.__new__(Vertex)
    result._v = v
    return result

def vertex_from_name(s):
    """Interpret the 'A1'-style notation for a point.

    Returns a Vertex.

    Raises ValueError if the string is invalid.

    Permits upper and lower case. Ignores surrounding whitespace.

    """
    cdef vertex_t v = vertex_from_alnum(string_from_cstr(s))
    if not v.is_on_board():
        raise ValueError
    return _vertex_from_vertex_t(v)


class IllegalMove(StandardError):
    """Exception used to indicate an illegal move."""

PLAY_OK = play_ok
PLAY_SUICIDE = play_suicide
PLAY_SS_SUICIDE = play_ss_suicide
PLAY_KO = play_ko

cdef class Board:
    """A go game position.

    This maintains the following state:
     - the colour of each point
     - the point (and player) which is forbidden due to simple ko, if any
     - the komi

    Initial state is empty with komi 7.5. Only fractional komi is supported.

    """

    cdef board_t *thisptr
    cdef bool_ board_is_borrowed
    cdef object board_owner

    def __cinit__(self):
        self.thisptr = new_board()
        self.board_is_borrowed = False
        self.board_owner = None

    def __dealloc__(self):
        if not self.board_is_borrowed:
            del_board(self.thisptr)

    cdef _borrow_board(self, object owner, board_t *b):
        """Set the managed board_t to one owned by a different object.

        b     -- board to borrow
        owner -- Python object which owns b

        This Board will keep a reference to 'owner', so that the board won't be
        freed as long as it's been borrowed.

        """
        assert not self.board_is_borrowed
        self.board_is_borrowed = True
        del_board(self.thisptr)
        self.thisptr = b
        self.board_owner = owner

    def point(self, Vertex v):
        """Return the status of the specified vertex.

        Returns BLACK, WHITE, or EMPTY.

        """
        return self.thisptr.color_at.get(v._v)

    def load_board(self, Board other not None):
        """Replace the state with the state from another board.

        (In libego, Board::load() is overloaded. This is the variant which takes
         a board_t* parameter. From Python code, using Board.copy() is probably
         more natural)

        """
        self.thisptr.load(other.thisptr)

    def copy(self):
        """Return a new Board with the same state as this one."""
        cdef Board result = Board()
        result.thisptr.load(self.thisptr)
        return result

    property hash:
        """The hash of the board position (read-only).

        This is a 64-bit unsigned integer.

        It represents the configuration of stones on the board. It doesn't
        depend on the ko-forbidden-point, the player who moved last, or the
        komi.

        Don't confuse this with the Python hash (that is based on the Board's
        identity).

        """
        def __get__(self):
            return self.thisptr.hash.hash

    property empty_v_cnt:
        """Number of empty points."""
        def __get__(self):
            return self.thisptr.empty_v_cnt

    property empty_vertices:
        """Describe the board's empty points.

        Returns a list of Vertices.

        """
        def __get__(self):
            cdef int i
            return [_vertex_from_vertex_t(self.thisptr.empty_v[i])
                    for i from 0 <= i < self.thisptr.empty_v_cnt]

    def get_empty_vertex(self, int i):
        """Return empty_vertices[i]."""
        if not 0 <= i < self.thisptr.empty_v_cnt:
            raise ValueError("out of range")
        return _vertex_from_vertex_t(self.thisptr.empty_v[i])

    def clear(self):
        """Clear the board position.

        This also resets the komi and clears the ko status.

        """
        self.thisptr.clear()

    def play_internal(self, int player, Vertex v not None):
        """Play a single non-pass move, without checking.

        (This is a simple wrapper for play_no_pass)

        This doesn't check whether the player value is in range; it will cause
        undefined behaviour if it is not.

        This doesn't check whether the point is occupied. The board will become
        inconsistent if it is.

        If the move is illegal due to simple ko, does not alter the board state
        and returns PLAY_KO.

        If the move is a single-stone suicide, returns PLAY_SS_SUICIDE.

        If the move is a multiple-stone suicide, plays the move and returns
        PLAY_SUICIDE.

        """
        return self.thisptr.play_no_pass(player_from_int(player), v._v)

    def play(self, int player, Vertex v not None):
        """Play a single non-pass move.

        player   -- BLACK or WHITE
        v        -- Vertex

        Raises IllegalMove if the specified point is occupied.

        Raises IllegalMove if the move is illegal due to simple ko, or if it is
        a single-stone suicide. Permits multiple-stone suicide.

        Raises ValueError if the player is not valid.

        """
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        if self.thisptr.color_at.get(v._v) != empty_color:
            raise IllegalMove("point is occupied")
        cdef int play_ret = self.thisptr.play_no_pass(pl, v._v)
        if play_ret == play_ko:
            raise IllegalMove("ko")
        if play_ret == play_ss_suicide:
            raise IllegalMove("single-stone suicide")

    def is_eyelike(self, int player, Vertex v not None):
        """Test whether a point is 'eyelike' for the specified player.

        This returns true exactly when
         - all its orthogonal neighbours are the player's colour, and
         - at most one diagonal neighbour is the opponent's colour
           (at most zero if the point is on the edge)

        Note that this does not check that the point itself is empty.

        """
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        return self.thisptr.is_eyelike(pl, v._v)

    def set_komi(self, float fkomi):
        """Set the komi."""
        self.thisptr.set_komi(fkomi)

    def get_komi(self):
        """Return the komi (as a float)."""
        return self.thisptr.get_komi()

    def score(self):
        """Return the score of a final position.

        Returns black's winning margin as an integer (0 means a half-point win
        for white).

        As I understand it, this works by counting stones and one-point 'eyes',
        so it should be used only when a game has progressed that far.

        """
        return self.thisptr.score()

    def approx_score(self):
        """Return a rough approximation to the score.

        This just counts the number of stones of each colour on the board, and
        adjusts for komi.

        It uses the same scale as score().

        """
        return self.thisptr.approx_score()

    def winner(self):
        """Return the winner as determined by score().

        Returns BLACK or WHITE.

        """
        return self.thisptr.winner().get_idx()

    def approx_winner(self):
        """Return the winner as determined by approx_score().

        Returns BLACK or WHITE.

        """
        return self.thisptr.approx_winner().get_idx()

    def to_string(self, Vertex mark_vertex=None):
        """Return an ascii-art representation of the board position.

        You can pass mark_vertex to indicate a point to be highlighted.

        """
        cdef string s
        if mark_vertex is None:
            s = self.thisptr.to_string()
        else:
            s = self.thisptr.to_string_with_mark(mark_vertex._v)
        return s.c_str()

    def load(self, s):
        """Load a board position from a string.

        Raises ValueError if the string is ill-formed. In this case, the board
        is cleared.

        If the board position is well-formed but invalid (ie, it has a group
        with no liberties), this writes a message to stderr and terminates the
        process (!).

        """
        cdef string src = string_from_cstr(s)
        if not load_board_from_string(self.thisptr[0], src):
            raise ValueError("bad board format")


PLAYOUT_OK = playout_ok
PLAYOUT_MERCY = playout_mercy
PLAYOUT_TOO_LONG = playout_too_long

def set_playout_seed(unsigned long i):
    """Seed the random number generator used for playouts.

    The same generator is shared by all playout objects.

    """
    set_playout_seed_impl(i)

def playout_rand_int(int max):
    """Return a random integer in range(0, max)

    This uses the random number generator used for playouts.

    """
    return playout_rand_int_impl(max)

cdef class Playout:
    """Support for random playouts.

    Instantiate with a board and the player who will play first.

    The playout methods modify this board.

    """

    cdef simple_playout_t *thisptr
    # This reference guarantees that the Board is not deleted before the Playout
    cdef Board _board

    def __cinit__(self, Board bd, int player):
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        self.thisptr = new_simple_playout(bd.thisptr, pl)
        self._board = bd

    def __dealloc__(self):
        del_simple_playout(self.thisptr)

    property board:
        """The board associated with this Playout (read-only)."""
        def __get__(self):
            return self._board

    property act_player:
        """The player who will play the next move in the playout."""
        def __get__(self):
            return self.thisptr.act_player.get_idx()
        def __set__(self, int player):
            cdef player_t pl = player_from_int(player)
            if not pl.in_range():
                raise ValueError("invalid player")
            self.thisptr.act_player = pl

    def get_last_play(self, int player):
        """Return the last play by the specified player.

        Returns Any if the player has not yet moved

        Returns Pass if the player passed.

        Otherwise returns a Vertex.

        """
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        cdef vertex_t v = self.thisptr.last_v.get(pl)
        if v.eq(vertex_pass):
            return Pass
        if v.eq(vertex_any):
            return Any
        cdef Vertex result = Vertex.__new__(Vertex)
        result._v = v
        return result

    def play_one(self):
        """Play a single move in the playout.

        This does not change the active player.

        The move is chosen at random, with a nearly flat distribution.

        Points which pass is_eyelike() are not chosen. Single-stone suicides are
        not chosen. The simple ko rule is enforced (but not superko).

        If no such move is available, a pass is played (this has no effect on
        the board, but it can be detected by get_last_play()).

        """
        self.thisptr.play_one()

    def run(self):
        """Play a playout to its conclusion.

        This will continue play until
         - the players pass in succession
         - max_playout_length moves have been played, or
         - the mercy rule fires

        The return value is a PLAYOUT_... constant which indicates which
        of these things happened.

        """
        return self.thisptr.run()


def run_simple_playout_benchmark(Board start_board not None,
                                 int playout_count,
                                 int first_player):
    """Run the specified number of playouts and report on the speed.

    Returns a string describing the results.

    The timings are based on CPU user time as returned by getrusage(3).

    """
    cdef player_t pl = player_from_int(first_player)
    if not pl.in_range():
        raise ValueError("invalid player")
    cdef ostringstream dst
    _run_simple_playout_benchmark(
        start_board.thisptr, playout_count, pl, dst)
    return dst.str().c_str()


cdef class StackBoard:
    """A sequence of board positions.

    This maintains a stack of at least one board position. When it is first
    instantiated, it contains one, empty, board.

    The board at the top of the stack is the _current board_.

    """
    cdef stack_board_t *thisptr

    def __cinit__(self):
        self.thisptr = new_stack_board()

    def __dealloc__(self):
        del_stack_board(self.thisptr)

    def act_board(self):
        """Retrieve the current board.

        Returns a Board.

        If you call this multiple times at the same 'level' of the stack, it
        will return distinct Python Board objects which have the same underlying
        libego board.

        """
        cdef Board result = Board()
        result._borrow_board(self, self.thisptr.act_board())
        return result

    def try_play(self, int player, object play):
        """Extend the stack by playing a move.

        play -- a Vertex, or Pass or Resign

        If the move is legal, this adds a new board to the top of the stack,
        which differs from the previous current board by the move played (if the
        move is Pass or Resign, the new board is identical to the previous
        current board).

        If the move is illegal, the stack is unchanged.

        Both single and multiple stone suicide are considered illegal.

        When 'play' is a vertex, this checks both the simple ko rule and the
        (positionsl) superko rule (using the board's hash for the latter).

        """
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        cdef vertex_t _v
        cdef Vertex v
        if play is Pass:
            _v = vertex_pass
        elif play is Resign:
            _v = vertex_resign
        elif play is None:
            raise TypeError("vertex required")
        else:
            v = play
            _v = v._v
        if not self.thisptr.try_play(pl, _v):
            raise IllegalMove

    def try_undo(self):
        """Discard the board from the top of the stack.

        Raises IndexError if there is only one board in the stack to start with.

        """
        if not self.thisptr.try_undo():
            raise IndexError("can't undo")

    def is_legal(self, int player, object play):
        """Check if a play would be legal.

        play -- as for try_play

        Returns a bool.

        See try_play() for the rules used.

        """
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        cdef vertex_t _v
        cdef Vertex v
        if play is Pass:
            _v = vertex_pass
        elif play is Resign:
            _v = vertex_resign
        elif play is None:
            raise TypeError("vertex required")
        else:
            v = play
            _v = v._v
        return bool(self.thisptr.is_legal(pl, _v))

    def is_hash_repeated(self):
        """Check if the current board position is a repeat.

        That is, checks whether the hash of the current board position matches
        the hash of any previous position in the stack.

        """
        return bool(self.thisptr.is_hash_repeated())

    def set_komi(self, float fkomi):
        """Set the komi for all boards in the stack."""
        self.thisptr.set_komi(fkomi)

    def load(self, s):
        """Load a board position from a string and replace the current stack.

        If this succeeds, the stack will then contain a single board with the
        given position.

        Raises ValueError if the string is ill-formed. In this case, the stack
        is unchanged.

        If the board position is well-formed but invalid (ie, it has a group
        with no liberties), this writes a message to stderr and terminates the
        process (!).

        """
        cdef string src = string_from_cstr(s)
        if not load_stack_board_from_string(self.thisptr[0], src):
            raise ValueError("bad board format")

cdef class Uct:
    """Use UCT with thin random playouts to choose a move.

    Instantiate with a StackBoard.

    """
    cdef uct_t *thisptr
    # This reference guarantees that the stack_board is not deleted before the
    # Uct object.
    cdef object _stack_board

    def __cinit__(self, StackBoard sb not None):
        self.thisptr = new_uct(sb.thisptr[0])
        self._stack_board = sb

    def __dealloc__(self):
        del_uct(self.thisptr)

    def genmove(self, int player):
        """Generate a move for the specified player.

        Returns a Vertex, or Pass or Resign.

        This generates a move for the current position of the StackBoard; the
        previous positions are relevant for superko.

        It doesn't play the move onto the stack.

        It produces some diagnostic output on stderr.

        It might take several seconds (see config.cpp for the parameters).

        """
        cdef player_t pl = player_from_int(player)
        if not pl.in_range():
            raise ValueError("invalid player")
        cdef vertex_t v = self.thisptr.genmove(pl)
        if v.eq(vertex_pass):
            return Pass
        if v.eq(vertex_any):
            return Any
        cdef Vertex result = Vertex.__new__(Vertex)
        result._v = v
        return result

cdef class GtpEngine:
    """Base class for GTP engines."""
    cdef gtp_engine_t *thisptr

cdef class GtpStaticCommands(GtpEngine):
    """Implement trivial GTP commands.

    This supports commands whose response is a fixed string.

    This is a GTP engine which can be registered with a Gtp object.

    """
    def __cinit__(self):
        self.thisptr = <gtp_engine_t*>new_gtp_static_commands()

    def __dealloc__(self):
        del_gtp_static_commands(<gtp_static_commands_t*>self.thisptr)

    def add(self, command, response):
        """Register a command and its response."""
        cdef string command_ = string_from_cstr(command)
        cdef string response_ = string_from_cstr(response)
        (<gtp_static_commands_t*>(self.thisptr)).add(command_, response_)

cdef class GtpBoard(GtpEngine):
    """Implement board-related commands.

    Instantiate with a StackBoard. The moves played will be recorded on and
    checked against its stack.

    This is a GTP engine which can be registered with a Gtp object.

    It implements the following commands:
      boardsize
      clear_board
      komi
      load_position
      play
      undo
      showboard
      playout_benchmark

    """
    # This reference guarantees that the stack_board is not deleted before the
    # GtpBoard object.
    cdef object _stack_board

    def __cinit__(self, StackBoard sb not None):
        self.thisptr = <gtp_engine_t*>new_gtp_board(sb.thisptr[0])
        self._stack_board = sb

    def __dealloc__(self):
        del_gtp_board(<gtp_board_t*>self.thisptr)

cdef class GtpGenmove(GtpEngine):
    """Implement genmove, using uct.

    Instantiate with a StackBoard and a Uct object.

    genmove will choose a move and play it on the stack.

    Normally you would register both a GtpBoard and a GtpGenmove, using the same
    StackBoard.

    """
    # These references are to stop the objects being deleted too early
    cdef object _stack_board
    cdef object _uct

    def __cinit__(self, StackBoard sb not None, Uct uct not None):
        self.thisptr = <gtp_engine_t*>new_gtp_genmove(sb.thisptr[0],
                                                      uct.thisptr[0])
        self._stack_board = sb
        self._uct = uct

    def __dealloc__(self):
        del_gtp_genmove(<gtp_genmove_t*>self.thisptr)

class PyGtpError(StandardError):
    """Invalid parameters to a Python GTP command."""

class PyGtpSyntaxError(StandardError):
    """Invalid parameters to a Python GTP command."""

cdef class GtpPy(GtpEngine):
    """Implement GTP commands from Python.

    To use this, inherit from this class and implement get_command_names and
    exec_command.

    get_command_names should return a sequence of strings naming the GTP
    commands that the object supports.

    If get_command_names raises an exception, the engine will decline to
    register any commands.

    exec_command is passed two arguments: the command name as a string, and the
    parameters as a list of strings.

    exec_command should return the GTP response as a string. Alternatively, it
    can raise PyGtpError or PyGtpSyntaxError to cause a GTP error response. If
    you raise PyGtpError with a string parameter, it will be included in the
    response.

    If any other exception is raised from exec_command, a description is printed
    to sys.stderr (you can alter this behaviour using sys.excepthook) and a GTP
    error response saying 'internal error from Python code' is returned.

    You must ensure that the response (including any PyGtpError parameter)
    doesn't contain two consecutive newlines (otherwise you'll mess up the GTP
    synchronisation).

    """
    cdef object PyGtpError_
    cdef object PyGtpSyntaxError_

    def __cinit__(self, *args):
        self.thisptr = <gtp_engine_t*>new_gtp_py(
            self, PyGtpError, PyGtpSyntaxError)
        self.PyGtpError_ = PyGtpError
        self.PyGtpSyntaxError_ = PyGtpSyntaxError

    def __dealloc__(self):
        del_gtp_py(<gtp_py_t*>self.thisptr)

    def get_command_names(self):
        raise NotImplementedError

    def exec_command(self, cmd, params):
        raise NotImplementedError


cdef class Gtp:
    """Speak the GTP protocol.

    This class implements some basic commands itself (including list_commands,
    known_command, quit, and run_gtp_file).

    Use register_engine to implement more commands.

    """
    cdef gtp_t *thisptr

    def __cinit__(self):
        self.thisptr = new_gtp()

    def __dealloc__(self):
        del_gtp(self.thisptr)

    def register_engine(self, GtpEngine engine not None):
        """Register an implementation of more commands.

        engine -- a concrete instance of a GtpEngine subclass.

        """
        self.thisptr.register_engine(engine.thisptr[0])

    def run_loop(self):
        """Read and respond to GTP commands.

        This reads from stdin and writes to stdout.

        It terminates at end of file on stdin, or when the 'quit' command is
        entered.

        Other commands should be registered with gtp.register_engine().

        """
        self.thisptr.run_loop()

